;;; Thread switching utility library.  Linux i386 family nasm.  Tested with
;;; gcc, though should work with other compilers.

;;; Author: T W Bennet, bennet@mc.edu
;;; You may use the code under the terms of the GNU Public License.  Here's the
;;; recommended statement:
;;; 
;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 2 of the License, or
;;; (at your option) any later version.
;;; 
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

;;; Global table for linking as shared library.
	extern	_GLOBAL_OFFSET_TABLE_
	global	regsave:function
	global	regrest:function
	global	restack:function
	global	_doabt:function
	global	setcleanup:function
	global	bareexit:function
	
	extern	puts
	extern	abort

	section	.data		

	;; cleanup method pointer
cleanup:
	dd	_doabt
	;; Error msg.
message:
	db	'Thread returned from topmost stack frame.', 10, 0

	section	.text

;;; Some critical offsets for the register save block.
o_esp	equ	16
o_ebx	equ	20
o_eax	equ	32
o_flg	equ	36
o_start	equ	40
o_xtr	equ	40
;;; Total size is 44, since we need an extra for the restore.
	
;;; === int regsave(unsigned int area[11]) ===
;;; This saves all the GP registers into a block of memory.  The saved state
;;; is adjusted so that restoring the state is just ready to return from
;;; this function.
regsave:
	push	ebx
	push	ecx

	;; Save everything but epc
	mov	ebx, [esp+12]	; Get the save area.
	mov	[ebx], esp	; Put the current stack ptr where epc goes.
	lea	esp, [ebx+o_start] ; Set stack ptr to save in save area.
	pushf			; Push flags.
	pushad			; Push the other stuff.
	mov	esp, [ebx]	; Restore the stack pointer.

	;; Fix stuff up in the save area.
	mov	ecx, [esp+4]	; Get original ebx value
	mov	[ebx+o_ebx], ecx; Place it in the save area.
	lea	ecx, [esp+8]	; Stack ready to return.
	mov	[ebx+o_esp], ecx; Place in save area.
	mov	ecx, [esp+8]	; Our return address.
	mov	[ebx], ecx	; Place in save area.

	;; Clean up and leave.
	pop	ecx
	pop	ebx
	mov	eax,0
	ret

		
;;; === regrest(unsigned int area[11], int retval) ===
;;; This restores a register record created by regsave.  As such, it will
;;; transfer control so that the regsave call which saved the registers will
;;; return again.  However, it will have the return value specified here,
;;; rather than zero.  Note: The return value trick is accomplished by saving
;;; the specified value into the state record before restoring it, so this
;;; record is actually modified.
regrest:
	;;  Don't need to save registers, since I'm trashing them anyway.
	mov	eax, [esp+4]	; Get restore area.
	mov	ebx, [esp+8]	; Move re-return value to
	mov	[eax+o_eax],ebx	;    saved state.
	mov	ecx, [eax+o_esp]; Get the pointer for the stack to reactivate.
	mov	edx, [eax]	; Move the return value from the saved reg
	mov	[ecx],edx	;    state to the stack being reactivated.
	mov	[eax+o_xtr], ecx; Store new stack ptr at end of save area.
	lea	esp, [eax+4]	; Set the stack ptr to the restore area.
	popad			; Get the registers.
	popf			; That too.
	pop	esp		; We put it just where it belongs.
	ret			; And off we go.

;;; === restack(void *newstack, int size, int nframe,int nparm) ===
;;; Switche the stack to use the area indicated by newstack.  The function 
;;; first copies nframe frames plus nparm additional 32-bit words into the new
;;; stack area so the program can function using the new stack.  The stored
;;; ebp values are adjusted to point to the corresponding position in the
;;; new stack, and the last return pointer is replaced with a pointer to
;;; the cleanup routine as specified by setcleanup.  This allows a thread
;;; library to clean up threads that return past the creation point.  It
;;; then changes the hardware stack pointer so the CPU is using the new stack.
;;;
;;; The function assumes ebp is in use a frame pointer, and uses this to
;;; determine the amount of existing stack data to copy to the new stack.
;;; It makes sure the ebp and stack pointer are correctly related, and that
;;; the total amount to copy is less than the size of the stack area.  The
;;; function will return -2 if the ebp is less than the stack pointer, or
;;; -1 if the new stack area is too small.  It is a bad idea to call restack
;;; from a function which has large stack data structures, since these will
;;; be copied, and the new stack space must be large enough to contain them
;;; and also have some left.  On success, restack returns the amount of unused
;;; space remaining on the new stack area.  (The caller might want to abort
;;; the larger operation if this number is too small.)
;;; 
restack:
	;; Check that the frame pointer exceeds the stack pointer,
	;; which should be true if the frame pointer is in use.
	cmp	esp, ebp
	jl	.cont1
	mov	eax, -2
	ret

.cont1:
	;; Do the conventional frame pointer thing, and push registers
	;; for saving.
	push	ebp
	mov	ebp, esp
	push	ebx

	;; Chase the stack location of the ebp of the last frame to copy.
	mov	ebx, ebp	; Init scan pointer in ebx.
	mov	eax, [ebp+16]	; Init frame count.
.chase:	
	cmp	eax,0		; Done?
	jle	.chased

	mov	ebx, [ebx]	; Chase one frame pointer and dec count.
	sub	eax, 1

	jmp	.chase
.chased:

	;; Compute the number of bytes to move.  
	mov	eax, ebx	; The last (chased-down) ebp,
	sub	eax, ebp	;   Less ours, plus
	mov	ecx, [ebp+20]	;   Number of extra words,
	sal	ecx, 2		;      times 4 for bytes.
	add	eax, ecx	; Add 'em

	;; Make sure there's room
	cmp	eax, [ebp+12]	; Compare to area size argument
	jle	.cont2		; Okay.  Proceed.
	pop	ebx		; Restore regs
	pop	ebp
	mov	eax, -1		; Return -1
	ret

.cont2:
	;; Make some stack space.
	pop	ebx		; We're going to move this.
	sub	esp, 4		; A useful extra slot.
	push	ebx		; A much nicer spot.
	push	esi		; Need these.
	push	edi
	
	;; Compute the return value here.  We'll use it later.
	mov	edx, [ebp+12]
	sub	edx, eax

	;; Move the size for the move instr.
	mov	ecx, eax

	;; Compute the dest address, and save it on the stack.
	mov	edi, [ebp+8]
	add	edi, [ebp+12]
	sub	edi, ecx
	mov	[ebp-4], edi

	;; Compute the source address.
	lea	esi, [ebp+4]
	
	;; Find the difference between source and dest, which we will
	;; need later.
	mov	eax, edi
	sub	eax, esi
	
	;; Adjust the count for doublewords and perform the copy.
	sar	ecx,2
	cld
	rep	movsd

	;; Scan the new stack area to adjust ebps stored there.
	mov	edi, [ebp]	; Ptr to first ebp in orig. of copied area.
	add	edi, eax	; Shift to same point in copy.
	mov	ecx, [ebp+16]	; Init frame count.
.fixfr:	
	cmp	ecx,0		; Done?
	jle	.frfixed

	mov	ebx, [edi]	; Get the old frame pointer value.
	add	ebx, eax	; Adjust it
	mov	[edi], ebx	; Save back.
	xchg	edi, ebx	; Proceed.  Old edi value in ebx after exit.
	sub	ecx, 1

	jmp	.fixfr

.frfixed:
	;; Get the GOT for shared libraries.  Straight from the doc.
	call	.gg1
.gg1:	pop	esi
	add	esi, _GLOBAL_OFFSET_TABLE_+$$-.gg1 wrt ..gotpc

	;; Change the caller's return value to transfer to the cleanup
	;; vector.
	mov	esi, [esi+cleanup wrt ..gotoff]; Get the value to store.
	mov	[ebx+4], esi 	; Store mod return in the new stack.

	;; Create a modified ebp to point to the new stack area.
	mov	ebp, [ebp]
	add	ebp, eax
	;; ebp is now "restored" to the corresponding location on the
	;; new stack.  

	;; Return value.
	mov	eax, edx

	;; Restore registers
	pop	edi
	pop	esi
	pop	ebx
	
	;; This will change to the new stack pointer,  If I didn't screw up,
	;; it will point the correct return value.  The documentation from
	;; Intel says it will not increment after placing the value in esp.
	;; Hopefully, they're right.
	pop	esp

	ret

;;; Default cleanup vector.  Abort the program.
_doabt:
	push	ebp
	mov	ebp, esp
	
	call	.gg3
.gg3:	pop	ebx
	add	ebx, _GLOBAL_OFFSET_TABLE_+$$-.gg3 wrt ..gotpc
	lea	eax, [ebx+message wrt ..gotoff]
	push	eax
	call	puts
	add	esp, 4
	call	abort WRT ..plt

;;; === void (* setcleanup(void (*)(void)))(void) ===
;;; When rethread initializes a new stack, it replaces the last return value
;;; with a cleanup routine that will get invoked if a thread tries to return
;;; from its shallowest frame.  By default, that the _doabt function above,
;;; which prints a message and aborts using calls from the C library.  This
;;; function specified some other callback to use for this purpose.  Of course,
;;; the change applies only to stacks initialized after this call is made.
setcleanup:
	;;  Get got.
	call	.gg2
.gg2:	pop	edx
	add	edx, _GLOBAL_OFFSET_TABLE_+$$-.gg2 wrt ..gotpc

	mov	ecx, [esp+4]
	mov	eax, [edx+cleanup wrt ..gotoff]
	mov	[edx+cleanup wrt ..gotoff], ecx
	ret


;; http://www.geocities.com/SiliconValley/Park/3230/x86asm/asml1013.html
;; http://www.technocage.com/~ray/notespage.jsp?pageName=nasmexamples&pageTitle=NASM+Examples
;; http://blogs.msdn.com/joelpob/archive/2004/04/21/117664.aspx
;; http://nasm.sourceforge.net/doc/html/nasmdoc0.html
